@@ -25,9 +25,10 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
-#include <sys/psx_syscall.h>
 #include <sys/syscall.h>
 
+#include "psx_syscall.h"
+
 /*
  * psx_load_syscalls() is weakly defined so we can have it overridden
  * by libpsx if it is linked. Specifically, when libcap calls
@@ -89,6 +90,7 @@ static struct psx_tracker_s {
     } cmd;
 
     struct sigaction sig_action;
+    struct sigaction chained_action;
     registered_thread_t *root;
 } psx_tracker;
 
@@ -123,6 +125,9 @@ static void psx_posix_syscall_actor(int signum, siginfo_t *info, void *ignore) {
     /* bail early if this isn't something we recognize */
     if (signum != psx_tracker.psx_sig || !psx_tracker.cmd.active ||
 	info == NULL || info->si_code != SI_TKILL || info->si_pid != getpid()) {
+	if (psx_tracker.chained_action.sa_sigaction != 0) {
+	    psx_tracker.chained_action.sa_sigaction(signum, info, ignore);
+	}
 	return;
     }
 
@@ -173,6 +178,34 @@ int __wrap_pthread_create(pthread_t *thread, const pthread_attr_t *attr,
 extern int __real_pthread_create(pthread_t *thread, const pthread_attr_t *attr,
 				 void *(*start_routine) (void *), void *arg);
 
+/*
+ * psx_confirm_sigaction reconfirms that the psx handler is the first
+ * handler to respond to the psx signal. It assumes that
+ * psx_tracker.psx_sig has been set.
+ */
+static void psx_confirm_sigaction(void) {
+    sigset_t mask, orig;
+    struct sigaction existing_sa;
+
+    /*
+     * Block interrupts while potentially rewriting the handler.
+     */
+    sigemptyset(&mask);
+    sigaddset(&mask, psx_tracker.psx_sig);
+    sigprocmask(SIG_BLOCK, &mask, &orig);
+
+    sigaction(psx_tracker.psx_sig, NULL, &existing_sa);
+    if (existing_sa.sa_sigaction != psx_posix_syscall_actor) {
+	memcpy(&psx_tracker.chained_action, &existing_sa, sizeof(struct sigaction));
+	psx_tracker.sig_action.sa_sigaction = psx_posix_syscall_actor;
+	sigemptyset(&psx_tracker.sig_action.sa_mask);
+	psx_tracker.sig_action.sa_flags = SA_SIGINFO | SA_ONSTACK | SA_RESTART;
+	sigaction(psx_tracker.psx_sig, &psx_tracker.sig_action, NULL);
+    }
+
+    sigprocmask(SIG_SETMASK, &orig, NULL);
+}
+
 /*
  * psx_syscall_start initializes the subsystem including initializing
  * the mutex.
@@ -184,15 +217,17 @@ static void psx_syscall_start(void) {
     pthread_atfork(_psx_prepare_fork, _psx_fork_completed, _psx_forked_child);
 
     /*
-     * glibc nptl picks from the SIGRTMIN end, so we pick from the
-     * SIGRTMAX end
+     * All sorts of things are assumed by Linux and glibc and/or musl
+     * about signal handlers and which can be blocked. Go has its own
+     * idiosyncrasies too. We tried SIGRTMAX until
+     *
+     *   https://bugzilla.kernel.org/show_bug.cgi?id=210533
+     *
+     * Our current strategy is to aggressively intercept SIGSYS.
      */
-    psx_tracker.psx_sig = SIGRTMAX;
-    psx_tracker.sig_action.sa_sigaction = psx_posix_syscall_actor;
-    sigemptyset(&psx_tracker.sig_action.sa_mask);
-    psx_tracker.sig_action.sa_flags = SA_SIGINFO | SA_RESTART;;
-    sigaction(psx_tracker.psx_sig, &psx_tracker.sig_action, NULL);
+    psx_tracker.psx_sig = SIGSYS;
 
+    psx_confirm_sigaction();
     psx_do_registration(); // register the main thread.
 
     psx_tracker.initialized = 1;
@@ -201,7 +236,8 @@ static void psx_syscall_start(void) {
 /*
  * This is the only way this library globally locks. Note, this is not
  * to be confused with psx_sig (interrupt) blocking - which is
- * performed around thread creation.
+ * performed around thread creation and when the signal handler is
+ * being confirmed.
  */
 static void psx_lock(void)
 {
@@ -336,11 +372,45 @@ typedef struct {
  *    https://sourceware.org/bugzilla/show_bug.cgi?id=12889
  */
 static void _psx_exiting(void *node) {
+    /*
+     * Until we are in the _PSX_EXITING state, we must not block the
+     * psx_sig interrupt for this dying thread. That is, until this
+     * exiting thread can set ref->gone to 1, this dying thread is
+     * still participating in the psx syscall distribution.
+     *
+     * See https://github.com/golang/go/issues/42494 for a situation
+     * where this code is called with psx_tracker.psx_sig blocked.
+     */
+    sigset_t sigbit, orig_sigbits;
+    sigemptyset(&sigbit);
+    pthread_sigmask(SIG_UNBLOCK, &sigbit, &orig_sigbits);
+    sigaddset(&sigbit, psx_tracker.psx_sig);
+    pthread_sigmask(SIG_UNBLOCK, &sigbit, NULL);
+
+    /*
+     * With psx_tracker.psx_sig unblocked we can wait until this
+     * thread can enter the _PSX_EXITING state.
+     */
     psx_new_state(_PSX_IDLE, _PSX_EXITING);
+
+    /*
+     * We now indicate that this thread is no longer participating in
+     * the psx mechanism.
+     */
     registered_thread_t *ref = node;
     pthread_mutex_lock(&ref->mu);
     ref->gone = 1;
     pthread_mutex_unlock(&ref->mu);
+
+    /*
+     * At this point, we can restore the calling sigmask to whatever
+     * the caller thought was appropriate for a dying thread to have.
+     */
+    pthread_sigmask(SIG_SETMASK, &orig_sigbits, NULL);
+
+    /*
+     * Allow the rest of the psx system carry on as per normal.
+     */
     psx_new_state(_PSX_EXITING, _PSX_IDLE);
 }
 
@@ -419,16 +489,6 @@ int __wrap_pthread_create(pthread_t *thread, const pthread_attr_t *attr,
     return ret;
 }
 
-/*
- * psx_pthread_create is a wrapper for pthread_create() that registers
- * the newly created thread. If your threads are created already, they
- * can be individually registered with psx_register().
- */
-int psx_pthread_create(pthread_t *thread, const pthread_attr_t *attr,
-		       void *(*start_routine) (void *), void *arg) {
-    return __wrap_pthread_create(thread, attr, start_routine, arg);
-}
-
 /*
  * __psx_immediate_syscall does one syscall using the current
  * process.
@@ -442,9 +502,9 @@ static long int __psx_immediate_syscall(long int syscall_nr,
 
     if (count > 3) {
 	psx_tracker.cmd.six = 1;
-	psx_tracker.cmd.arg1 = arg[3];
-	psx_tracker.cmd.arg2 = count > 4 ? arg[4] : 0;
-	psx_tracker.cmd.arg3 = count > 5 ? arg[5] : 0;
+	psx_tracker.cmd.arg4 = arg[3];
+	psx_tracker.cmd.arg5 = count > 4 ? arg[4] : 0;
+	psx_tracker.cmd.arg6 = count > 5 ? arg[5] : 0;
 	return syscall(syscall_nr,
 		      psx_tracker.cmd.arg1,
 		      psx_tracker.cmd.arg2,
@@ -497,10 +557,11 @@ long int __psx_syscall(long int syscall_nr, ...) {
     }
 
     psx_new_state(_PSX_IDLE, _PSX_SETUP);
+    psx_confirm_sigaction();
 
     long int ret;
 
-    ret = __psx_immediate_syscall(syscall_nr, count, arg);;
+    ret = __psx_immediate_syscall(syscall_nr, count, arg);
     if (ret == -1 || !psx_tracker.initialized) {
 	psx_new_state(_PSX_SETUP, _PSX_IDLE);
 	goto defer;
